@@ -1,4 +1,4 @@
-/* $Id$ */
+/* $OpenBSD$ */
 
 /*
  * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -26,13 +26,14 @@
 #include <errno.h>
 #include <event.h>
 #include <fcntl.h>
-#include <pwd.h>
+#include <signal.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
 #include "tmux.h"
 
+int		client_flags;
 struct imsgbuf	client_ibuf;
 struct event	client_event;
 struct event	client_stdin;
@@ -51,9 +52,10 @@ enum msgtype	client_exittype;
 const char     *client_exitsession;
 int		client_attached;
 
+__dead void	client_exec(const char *);
 int		client_get_lock(char *);
-int		client_connect(char *, int);
-void		client_send_identify(int);
+int		client_connect(struct event_base *, char *, int);
+void		client_send_identify(const char *, int);
 int		client_write_one(enum msgtype, int, const void *, size_t);
 int		client_write_server(enum msgtype, const void *, size_t);
 void		client_update_event(void);
@@ -62,7 +64,7 @@ void		client_stdin_callback(int, short, void *);
 void		client_write(int, const char *, size_t);
 void		client_callback(int, short, void *);
 int		client_dispatch_attached(void);
-int		client_dispatch_wait(void *);
+int		client_dispatch_wait(void);
 const char     *client_exit_message(void);
 
 /*
@@ -78,25 +80,30 @@ client_get_lock(char *lockfile)
 
 	if ((lockfd = open(lockfile, O_WRONLY|O_CREAT, 0600)) == -1)
 		fatal("open failed");
+	log_debug("lock file is %s", lockfile);
 
-	if (lockf(lockfd, F_TLOCK, 0) == -1 && errno == EAGAIN) {
-		while (lockf(lockfd, F_LOCK, 0) == -1 && errno == EINTR)
+	if (flock(lockfd, LOCK_EX|LOCK_NB) == -1) {
+		log_debug("flock failed: %s", strerror(errno));
+		if (errno != EAGAIN)
+			return (lockfd);
+		while (flock(lockfd, LOCK_EX) == -1 && errno == EINTR)
 			/* nothing */;
 		close(lockfd);
 		return (-1);
 	}
+	log_debug("flock succeeded");
 
 	return (lockfd);
 }
 
 /* Connect client to server. */
 int
-client_connect(char *path, int start_server)
+client_connect(struct event_base *base, char *path, int start_server)
 {
 	struct sockaddr_un	sa;
 	size_t			size;
-	int			fd, lockfd;
-	char		       *lockfile;
+	int			fd, lockfd = -1, locked = 0;
+	char		       *lockfile = NULL;
 
 	memset(&sa, 0, sizeof sa);
 	sa.sun_family = AF_UNIX;
@@ -105,37 +112,60 @@ client_connect(char *path, int start_server)
 		errno = ENAMETOOLONG;
 		return (-1);
 	}
+	log_debug("socket is %s", path);
 
 retry:
 	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
 		fatal("socket failed");
 
-	if (connect(fd, (struct sockaddr *) &sa, SUN_LEN(&sa)) == -1) {
+	log_debug("trying connect");
+	if (connect(fd, (struct sockaddr *) &sa, sizeof(sa)) == -1) {
+		log_debug("connect failed: %s", strerror(errno));
 		if (errno != ECONNREFUSED && errno != ENOENT)
 			goto failed;
 		if (!start_server)
 			goto failed;
 		close(fd);
 
-		xasprintf(&lockfile, "%s.lock", path);
-		if ((lockfd = client_get_lock(lockfile)) == -1) {
-			free(lockfile);
+		if (!locked) {
+			xasprintf(&lockfile, "%s.lock", path);
+			if ((lockfd = client_get_lock(lockfile)) == -1) {
+				log_debug("didn't get lock");
+				free(lockfile);
+				goto retry;
+			}
+			log_debug("got lock");
+
+			/*
+			 * Always retry at least once, even if we got the lock,
+			 * because another client could have taken the lock,
+			 * started the server and released the lock between our
+			 * connect() and flock().
+			 */
+			locked = 1;
 			goto retry;
 		}
+
 		if (unlink(path) != 0 && errno != ENOENT) {
 			free(lockfile);
 			close(lockfd);
 			return (-1);
 		}
-		fd = server_start(lockfd, lockfile);
+		fd = server_start(base, lockfd, lockfile);
+	}
+
+	if (locked) {
 		free(lockfile);
 		close(lockfd);
 	}
-
 	setblocking(fd, 0);
 	return (fd);
 
 failed:
+	if (locked) {
+		free(lockfile);
+		close(lockfd);
+	}
 	close(fd);
 	return (-1);
 }
@@ -179,18 +209,22 @@ client_exit_message(void)
 
 /* Client main loop. */
 int
-client_main(int argc, char **argv, int flags)
+client_main(struct event_base *base, int argc, char **argv, int flags)
 {
 	struct cmd		*cmd;
 	struct cmd_list		*cmdlist;
 	struct msg_command_data	*data;
-	int			 cmdflags, fd, i;
+	int			 cmdflags, fd, i, cwd;
+	const char*              ttynam;
 	pid_t			 ppid;
 	enum msgtype		 msg;
 	char			*cause;
 	struct termios		 tio, saved_tio;
 	size_t			 size;
 
+	/* Save the flags. */
+	client_flags = flags;
+
 	/* Set up the initial command. */
 	cmdflags = 0;
 	if (shell_cmd != NULL) {
@@ -198,7 +232,7 @@ client_main(int argc, char **argv, int flags)
 		cmdflags = CMD_STARTSERVER;
 	} else if (argc == 0) {
 		msg = MSG_COMMAND;
-		cmdflags = CMD_STARTSERVER|CMD_CANTNEST;
+		cmdflags = CMD_STARTSERVER;
 	} else {
 		msg = MSG_COMMAND;
 
@@ -216,31 +250,51 @@ client_main(int argc, char **argv, int flags)
 		TAILQ_FOREACH(cmd, &cmdlist->list, qentry) {
 			if (cmd->entry->flags & CMD_STARTSERVER)
 				cmdflags |= CMD_STARTSERVER;
-			if (cmd->entry->flags & CMD_CANTNEST)
-				cmdflags |= CMD_CANTNEST;
 		}
 		cmd_list_free(cmdlist);
 	}
 
-	/*
-	 * Check if this could be a nested session, if the command can't nest:
-	 * if the socket path matches $TMUX, this is probably the same server.
-	 */
-	if (shell_cmd == NULL && environ_path != NULL &&
-	    (cmdflags & CMD_CANTNEST) &&
-	    strcmp(socket_path, environ_path) == 0) {
-		fprintf(stderr, "sessions should be nested with care, "
-		    "unset $TMUX to force\n");
+	/* Establish signal handlers. */
+	set_signals(client_signal);
+
+	/* Initialize the client socket and start the server. */
+	fd = client_connect(base, socket_path, cmdflags & CMD_STARTSERVER);
+	if (fd == -1) {
+		if (errno == ECONNREFUSED) {
+			fprintf(stderr, "no server running on %s\n",
+			    socket_path);
+		} else {
+			fprintf(stderr, "error connecting to %s (%s)\n",
+			    socket_path, strerror(errno));
+		}
 		return (1);
 	}
 
-	/* Initialise the client socket and start the server. */
-	fd = client_connect(socket_path, cmdflags & CMD_STARTSERVER);
-	if (fd == -1) {
-		fprintf(stderr, "failed to connect to server: %s\n",
-		    strerror(errno));
-		return (1);
-	}
+	/* Save these before pledge(). */
+	if ((cwd = open(".", O_RDONLY)) == -1)
+		cwd = open("/", O_RDONLY);
+	if ((ttynam = ttyname(STDIN_FILENO)) == NULL)
+		ttynam = "";
+
+#ifdef __OpenBSD__
+	/*
+	 * Drop privileges for client. "proc exec" is needed for -c and for
+	 * locking (which uses system(3)).
+	 *
+	 * "tty" is needed to restore termios(4) and also for some reason -CC
+	 * does not work properly without it (input is not recognised).
+	 *
+	 * "sendfd" is dropped later in client_dispatch_wait().
+	 */
+	if (pledge("stdio unix sendfd proc exec tty", NULL) != 0)
+		fatal("pledge failed");
+#endif
+
+	/* Free stuff that is not used in the client. */
+	options_free(&global_options);
+	options_free(&global_s_options);
+	options_free(&global_w_options);
+	environ_free(&global_environ);
 
 	/* Set process title, log and signals now this is the client. */
 #ifdef HAVE_SETPROCTITLE
@@ -250,13 +304,13 @@ client_main(int argc, char **argv, int flags)
 
 	/* Create imsg. */
 	imsg_init(&client_ibuf, fd);
-	event_set(&client_event, fd, EV_READ, client_callback, shell_cmd);
+	event_set(&client_event, fd, EV_READ, client_callback, NULL);
 
 	/* Create stdin handler. */
 	setblocking(STDIN_FILENO, 0);
 	event_set(&client_stdin, STDIN_FILENO, EV_READ|EV_PERSIST,
 	    client_stdin_callback, NULL);
-	if (flags & CLIENT_CONTROLCONTROL) {
+	if (client_flags & CLIENT_CONTROLCONTROL) {
 		if (tcgetattr(STDIN_FILENO, &saved_tio) != 0) {
 			fprintf(stderr, "tcgetattr failed: %s\n",
 			    strerror(errno));
@@ -276,11 +330,8 @@ client_main(int argc, char **argv, int flags)
 		tcsetattr(STDIN_FILENO, TCSANOW, &tio);
 	}
 
-	/* Establish signal handlers. */
-	set_signals(client_signal);
-
 	/* Send identify messages. */
-	client_send_identify(flags);
+	client_send_identify(ttynam, cwd); /* closes cwd */
 
 	/* Send first command. */
 	if (msg == MSG_COMMAND) {
@@ -292,7 +343,7 @@ client_main(int argc, char **argv, int flags)
 
 		/* Prepare command for server. */
 		data->argc = argc;
-		if (cmd_pack_argv(argc, argv, (char*)(data + 1), size) != 0) {
+		if (cmd_pack_argv(argc, argv, (char *)(data + 1), size) != 0) {
 			fprintf(stderr, "command too long\n");
 			free(data);
 			return (1);
@@ -315,13 +366,13 @@ client_main(int argc, char **argv, int flags)
 
 	/* Print the exit message, if any, and exit. */
 	if (client_attached) {
-		if (client_exitreason != CLIENT_EXIT_NONE && !login_shell)
+		if (client_exitreason != CLIENT_EXIT_NONE)
 			printf("[%s]\n", client_exit_message());
 
 		ppid = getppid();
 		if (client_exittype == MSG_DETACHKILL && ppid > 1)
 			kill(ppid, SIGHUP);
-	} else if (flags & CLIENT_CONTROLCONTROL) {
+	} else if (client_flags & CLIENT_CONTROLCONTROL) {
 		if (client_exitreason != CLIENT_EXIT_NONE)
 			printf("%%exit %s\n", client_exit_message());
 		else
@@ -335,11 +386,13 @@ client_main(int argc, char **argv, int flags)
 
 /* Send identify messages to server. */
 void
-client_send_identify(int flags)
+client_send_identify(const char *ttynam, int cwd)
 {
-	const char	*s;
+	const char	 *s;
 	char		**ss;
-	int		 fd;
+	size_t		  sslen;
+	int		  fd, flags = client_flags;
+	pid_t		  pid;
 
 	client_write_one(MSG_IDENTIFY_FLAGS, -1, &flags, sizeof flags);
 
@@ -347,24 +400,23 @@ client_send_identify(int flags)
 		s = "";
 	client_write_one(MSG_IDENTIFY_TERM, -1, s, strlen(s) + 1);
 
-	if ((s = ttyname(STDIN_FILENO)) == NULL)
-		s = "";
-	client_write_one(MSG_IDENTIFY_TTYNAME, -1, s, strlen(s) + 1);
-
-	if ((fd = open(".", O_RDONLY)) == -1)
-		fd = open("/", O_RDONLY);
-	client_write_one(MSG_IDENTIFY_CWD, fd, NULL, 0);
+	client_write_one(MSG_IDENTIFY_TTYNAME, -1, ttynam, strlen(ttynam) + 1);
+	client_write_one(MSG_IDENTIFY_CWD, cwd, NULL, 0);
 
 	if ((fd = dup(STDIN_FILENO)) == -1)
 		fatal("dup failed");
 	client_write_one(MSG_IDENTIFY_STDIN, fd, NULL, 0);
 
-	for (ss = environ; *ss != NULL; ss++)
-		client_write_one(MSG_IDENTIFY_ENVIRON, -1, *ss, strlen(*ss) + 1);
+	pid = getpid();
+	client_write_one(MSG_IDENTIFY_CLIENTPID, -1, &pid, sizeof pid);
+
+	for (ss = environ; *ss != NULL; ss++) {
+		sslen = strlen(*ss) + 1;
+		if (sslen <= MAX_IMSGSIZE - IMSG_HEADER_SIZE)
+			client_write_one(MSG_IDENTIFY_ENVIRON, -1, *ss, sslen);
+	}
 
 	client_write_one(MSG_IDENTIFY_DONE, -1, NULL, 0);
-
-	client_update_event();
 }
 
 /* Helper to send one message. */
@@ -374,7 +426,7 @@ client_write_one(enum msgtype type, int fd, const void *buf, size_t len)
 	int	retval;
 
 	retval = imsg_compose(&client_ibuf, type, PROTOCOL_VERSION, -1, fd,
-	    (void*)buf, len);
+	    (void *)buf, len);
 	if (retval != 1)
 		return (-1);
 	return (0);
@@ -402,27 +454,22 @@ client_update_event(void)
 	events = EV_READ;
 	if (client_ibuf.w.queued > 0)
 		events |= EV_WRITE;
-	event_set(
-	    &client_event, client_ibuf.fd, events, client_callback, shell_cmd);
+	event_set(&client_event, client_ibuf.fd, events, client_callback, NULL);
 	event_add(&client_event, NULL);
 }
 
 /* Callback to handle signals in the client. */
 void
-client_signal(int sig, unused short events, unused void *data)
+client_signal(int sig, unused short events, unused void *arg)
 {
 	struct sigaction sigact;
 	int		 status;
 
-	if (!client_attached) {
-		switch (sig) {
-		case SIGCHLD:
-			waitpid(WAIT_ANY, &status, WNOHANG);
-			break;
-		case SIGTERM:
+	if (sig == SIGCHLD)
+		waitpid(WAIT_ANY, &status, WNOHANG);
+	else if (!client_attached) {
+		if (sig == SIGTERM)
 			event_loopexit(NULL);
-			break;
-		}
 	} else {
 		switch (sig) {
 		case SIGHUP:
@@ -455,7 +502,7 @@ client_signal(int sig, unused short events, unused void *data)
 
 /* Callback for client imsg read events. */
 void
-client_callback(unused int fd, short events, void *data)
+client_callback(unused int fd, short events, unused void *arg)
 {
 	ssize_t	n;
 	int	retval;
@@ -466,7 +513,7 @@ client_callback(unused int fd, short events, void *data)
 		if (client_attached)
 			retval = client_dispatch_attached();
 		else
-			retval = client_dispatch_wait(data);
+			retval = client_dispatch_wait();
 		if (retval != 0) {
 			event_loopexit(NULL);
 			return;
@@ -474,7 +521,7 @@ client_callback(unused int fd, short events, void *data)
 	}
 
 	if (events & EV_WRITE) {
-		if (msgbuf_write(&client_ibuf.w) < 0 && errno != EAGAIN)
+		if (msgbuf_write(&client_ibuf.w) <= 0 && errno != EAGAIN)
 			goto lost_server;
 	}
 
@@ -489,7 +536,7 @@ lost_server:
 
 /* Callback for client stdin read events. */
 void
-client_stdin_callback(unused int fd, unused short events, unused void *data1)
+client_stdin_callback(unused int fd, unused short events, unused void *arg)
 {
 	struct msg_stdin_data	data;
 
@@ -521,9 +568,38 @@ client_write(int fd, const char *data, size_t size)
 	}
 }
 
+/* Run command in shell; used for -c. */
+__dead void
+client_exec(const char *shell)
+{
+	const char	*name, *ptr;
+	char		*argv0;
+
+	log_debug("shell %s, command %s", shell, shell_cmd);
+
+	ptr = strrchr(shell, '/');
+	if (ptr != NULL && *(ptr + 1) != '\0')
+		name = ptr + 1;
+	else
+		name = shell;
+	if (client_flags & CLIENT_LOGIN)
+		xasprintf(&argv0, "-%s", name);
+	else
+		xasprintf(&argv0, "%s", name);
+	setenv("SHELL", shell, 1);
+
+	setblocking(STDIN_FILENO, 1);
+	setblocking(STDOUT_FILENO, 1);
+	setblocking(STDERR_FILENO, 1);
+	closefrom(STDERR_FILENO + 1);
+
+	execl(shell, argv0, "-c", shell_cmd, (char *) NULL);
+	fatal("execl failed");
+}
+
 /* Dispatch imsgs when in wait state (before MSG_READY). */
 int
-client_dispatch_wait(void *data0)
+client_dispatch_wait(void)
 {
 	struct imsg		 imsg;
 	char			*data;
@@ -531,6 +607,21 @@ client_dispatch_wait(void *data0)
 	struct msg_stdout_data	 stdoutdata;
 	struct msg_stderr_data	 stderrdata;
 	int			 retval;
+#ifdef __OpenBSD__
+	static int		 pledge_applied;
+
+	/*
+	 * "sendfd" is no longer required once all of the identify messages
+	 * have been sent. We know the server won't send us anything until that
+	 * point (because we don't ask it to), so we can drop "sendfd" once we
+	 * get the first message from the server.
+	 */
+	if (!pledge_applied) {
+		if (pledge("stdio unix proc exec tty", NULL) != 0)
+			fatal("pledge failed");
+		pledge_applied = 1;
+	};
+#endif
 
 	for (;;) {
 		if ((n = imsg_get(&client_ibuf, &imsg)) == -1)
@@ -541,7 +632,7 @@ client_dispatch_wait(void *data0)
 		data = imsg.data;
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 
-		log_debug("got %d from server", imsg.hdr.type);
+		log_debug("got %u from server", imsg.hdr.type);
 		switch (imsg.hdr.type) {
 		case MSG_EXIT:
 		case MSG_SHUTDOWN:
@@ -588,7 +679,7 @@ client_dispatch_wait(void *data0)
 				fatalx("bad MSG_VERSION size");
 
 			fprintf(stderr, "protocol version mismatch "
-			    "(client %u, server %u)\n", PROTOCOL_VERSION,
+			    "(client %d, server %u)\n", PROTOCOL_VERSION,
 			    imsg.hdr.peerid);
 			client_exitval = 1;
 
@@ -599,7 +690,7 @@ client_dispatch_wait(void *data0)
 				fatalx("bad MSG_SHELL string");
 
 			clear_signals(0);
-			shell_exec(data, data0);
+			client_exec(data);
 			/* NOTREACHED */
 		case MSG_DETACH:
 		case MSG_DETACHKILL:
@@ -632,7 +723,7 @@ client_dispatch_attached(void)
 		data = imsg.data;
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 
-		log_debug("got %d from server", imsg.hdr.type);
+		log_debug("got %u from server", imsg.hdr.type);
 		switch (imsg.hdr.type) {
 		case MSG_DETACH:
 		case MSG_DETACHKILL:
