@@ -1,4 +1,4 @@
-/* $Id$ */
+/* $OpenBSD$ */
 
 /*
  * Copyright (c) 2009 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -26,14 +26,15 @@
  * Break pane off into a window.
  */
 
+#define BREAK_PANE_TEMPLATE "#{session_name}:#{window_index}.#{pane_index}"
+
 enum cmd_retval	 cmd_break_pane_exec(struct cmd *, struct cmd_q *);
 
 const struct cmd_entry cmd_break_pane_entry = {
 	"break-pane", "breakp",
-	"dPF:t:", 0, 0,
-	"[-dP] [-F format] " CMD_TARGET_PANE_USAGE,
+	"dPF:s:t:", 0, 0,
+	"[-dP] [-F format] " CMD_SRCDST_PANE_USAGE,
 	0,
-	NULL,
 	cmd_break_pane_exec
 };
 
@@ -42,42 +43,39 @@ cmd_break_pane_exec(struct cmd *self, struct cmd_q *cmdq)
 {
 	struct args		*args = self->args;
 	struct winlink		*wl;
-	struct session		*s;
+	struct session		*src_s;
+	struct session		*dst_s;
 	struct window_pane	*wp;
 	struct window		*w;
 	char			*name;
 	char			*cause;
-	int			 base_idx;
-	struct client		*c;
+	int			 idx;
 	struct format_tree	*ft;
 	const char		*template;
 	char			*cp;
 
-	if ((wl = cmd_find_pane(cmdq, args_get(args, 't'), &s, &wp)) == NULL)
+	wl = cmd_find_pane(cmdq, args_get(args, 's'), &src_s, &wp);
+	if (wl == NULL)
 		return (CMD_RETURN_ERROR);
+	if ((idx = cmd_find_index(cmdq, args_get(args, 't'), &dst_s)) == -2)
+		return (CMD_RETURN_ERROR);
+	if (idx != -1 && winlink_find_by_index(&dst_s->windows, idx) != NULL) {
+		cmdq_error(cmdq, "index %d already in use", idx);
+		return (CMD_RETURN_ERROR);
+	}
+	w = wl->window;
 
-	if (window_count_panes(wl->window) == 1) {
+	if (window_count_panes(w) == 1) {
 		cmdq_error(cmdq, "can't break with only one pane");
 		return (CMD_RETURN_ERROR);
 	}
-
-	w = wl->window;
 	server_unzoom_window(w);
 
 	TAILQ_REMOVE(&w->panes, wp, entry);
-	if (wp == w->active) {
-		w->active = w->last;
-		w->last = NULL;
-		if (w->active == NULL) {
-			w->active = TAILQ_PREV(wp, window_panes, entry);
-			if (w->active == NULL)
-				w->active = TAILQ_NEXT(wp, entry);
-		}
-	} else if (wp == w->last)
-		w->last = NULL;
+	window_lost_pane(w, wp);
 	layout_close_pane(wp);
 
-	w = wp->window = window_create1(s->sx, s->sy);
+	w = wp->window = window_create1(dst_s->sx, dst_s->sy);
 	TAILQ_INSERT_HEAD(&w->panes, wp, entry);
 	w->active = wp;
 	name = default_window_name(w);
@@ -85,24 +83,26 @@ cmd_break_pane_exec(struct cmd *self, struct cmd_q *cmdq)
 	free(name);
 	layout_init(w, wp);
 
-	base_idx = options_get_number(&s->options, "base-index");
-	wl = session_attach(s, w, -1 - base_idx, &cause); /* can't fail */
+	if (idx == -1)
+		idx = -1 - options_get_number(&dst_s->options, "base-index");
+	wl = session_attach(dst_s, w, idx, &cause); /* can't fail */
 	if (!args_has(self->args, 'd'))
-		session_select(s, wl->idx);
+		session_select(dst_s, wl->idx);
 
-	server_redraw_session(s);
-	server_status_session_group(s);
+	server_redraw_session(src_s);
+	if (src_s != dst_s)
+		server_redraw_session(dst_s);
+	server_status_session_group(src_s);
+	if (src_s != dst_s)
+		server_status_session_group(dst_s);
 
 	if (args_has(args, 'P')) {
 		if ((template = args_get(args, 'F')) == NULL)
 			template = BREAK_PANE_TEMPLATE;
 
 		ft = format_create();
-		if ((c = cmd_find_client(cmdq, NULL, 1)) != NULL)
-			format_client(ft, c);
-		format_session(ft, s);
-		format_winlink(ft, s, wl);
-		format_window_pane(ft, wp);
+		format_defaults(ft, cmd_find_client(cmdq, NULL, 1), dst_s, wl,
+		    wp);
 
 		cp = format_expand(ft, template);
 		cmdq_print(cmdq, "%s", cp);
