@@ -41,16 +41,16 @@ enum cmd_retval
 cmd_kill_session_exec(struct cmd *self, struct cmd_q *cmdq)
 {
 	struct args	*args = self->args;
-	struct session	*s, *s2, *s3;
+	struct session	*s, *sloop, *stmp;
 
 	if ((s = cmd_find_session(cmdq, args_get(args, 't'), 0)) == NULL)
 		return (CMD_RETURN_ERROR);
 
 	if (args_has(args, 'a')) {
-		RB_FOREACH_SAFE(s2, sessions, &sessions, s3) {
-			if (s != s2) {
-				server_destroy_session(s2);
-				session_destroy(s2);
+		RB_FOREACH_SAFE(sloop, sessions, &sessions, stmp) {
+			if (sloop != s) {
+				server_destroy_session(sloop);
+				session_destroy(sloop);
 			}
 		}
 	} else {
