@@ -1,4 +1,4 @@
-/* $Id$ */
+/* $OpenBSD$ */
 
 /*
  * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -33,7 +33,6 @@ const struct cmd_entry cmd_rename_window_entry = {
 	"t:", 1, 1,
 	CMD_TARGET_WINDOW_USAGE " new-name",
 	0,
-	NULL,
 	cmd_rename_window_exec
 };
 
