@@ -24,7 +24,7 @@
 #include "tmux.h"
 
 /*
- * Add, set, or append to a paste buffer.
+ * Add, set, append to or delete a paste buffer.
  */
 
 enum cmd_retval	 cmd_set_buffer_exec(struct cmd *, struct cmd_q *);
@@ -37,41 +37,52 @@ const struct cmd_entry cmd_set_buffer_entry = {
 	cmd_set_buffer_exec
 };
 
+const struct cmd_entry cmd_delete_buffer_entry = {
+	"delete-buffer", "deleteb",
+	"b:", 0, 0,
+	CMD_BUFFER_USAGE,
+	0,
+	cmd_set_buffer_exec
+};
+
 enum cmd_retval
 cmd_set_buffer_exec(struct cmd *self, struct cmd_q *cmdq)
 {
 	struct args		*args = self->args;
 	struct paste_buffer	*pb;
-	char			*pdata, *cause;
-	const char		*bufname;
-	size_t			 psize, newsize;
+	char			*bufdata, *cause;
+	const char		*bufname, *olddata;
+	size_t			 bufsize, newsize;
 
-	bufname = NULL;
+	bufname = args_get(args, 'b');
+	if (bufname == NULL)
+		pb = NULL;
+	else
+		pb = paste_get_name(bufname);
 
-	if (args_has(args, 'n')) {
-		if (args->argc > 0) {
-			cmdq_error(cmdq, "don't provide data with n flag");
+	if (self->entry == &cmd_delete_buffer_entry) {
+		if (pb == NULL)
+			pb = paste_get_top(&bufname);
+		if (pb == NULL) {
+			cmdq_error(cmdq, "no buffer");
 			return (CMD_RETURN_ERROR);
 		}
+		paste_free(pb);
+		return (CMD_RETURN_NORMAL);
+	}
 
-		if (args_has(args, 'b'))
-			bufname = args_get(args, 'b');
-
-		if (bufname == NULL) {
-			pb = paste_get_top();
-			if (pb == NULL) {
-				cmdq_error(cmdq, "no buffer");
-				return (CMD_RETURN_ERROR);
-			}
-			bufname = pb->name;
+	if (args_has(args, 'n')) {
+		if (pb == NULL)
+			pb = paste_get_top(&bufname);
+		if (pb == NULL) {
+			cmdq_error(cmdq, "no buffer");
+			return (CMD_RETURN_ERROR);
 		}
-
 		if (paste_rename(bufname, args_get(args, 'n'), &cause) != 0) {
 			cmdq_error(cmdq, "%s", cause);
 			free(cause);
 			return (CMD_RETURN_ERROR);
 		}
-
 		return (CMD_RETURN_NORMAL);
 	}
 
@@ -79,37 +90,25 @@ cmd_set_buffer_exec(struct cmd *self, struct cmd_q *cmdq)
 		cmdq_error(cmdq, "no data specified");
 		return (CMD_RETURN_ERROR);
 	}
-
-	psize = 0;
-	pdata = NULL;
-
-	pb = NULL;
-
 	if ((newsize = strlen(args->argv[0])) == 0)
 		return (CMD_RETURN_NORMAL);
 
-	if (args_has(args, 'b')) {
-		bufname = args_get(args, 'b');
-		pb = paste_get_name(bufname);
-	} else if (args_has(args, 'a')) {
-		pb = paste_get_top();
-		if (pb != NULL)
-			bufname = pb->name;
-	}
+	bufsize = 0;
+	bufdata = NULL;
 
 	if (args_has(args, 'a') && pb != NULL) {
-		psize = pb->size;
-		pdata = xmalloc(psize);
-		memcpy(pdata, pb->data, psize);
+		olddata = paste_buffer_data(pb, &bufsize);
+		bufdata = xmalloc(bufsize);
+		memcpy(bufdata, olddata, bufsize);
 	}
 
-	pdata = xrealloc(pdata, psize + newsize);
-	memcpy(pdata + psize, args->argv[0], newsize);
-	psize += newsize;
+	bufdata = xrealloc(bufdata, bufsize + newsize);
+	memcpy(bufdata + bufsize, args->argv[0], newsize);
+	bufsize += newsize;
 
-	if (paste_set(pdata, psize, bufname, &cause) != 0) {
+	if (paste_set(bufdata, bufsize, bufname, &cause) != 0) {
 		cmdq_error(cmdq, "%s", cause);
-		free(pdata);
+		free(bufdata);
 		free(cause);
 		return (CMD_RETURN_ERROR);
 	}
