@@ -146,6 +146,7 @@ cmd_split_window_exec(struct cmd *self, struct cmd_q *cmdq)
 		goto error;
 	}
 	new_wp = window_add_pane(w, hlimit);
+	layout_assign_pane(lc, new_wp);
 
 	path = NULL;
 	if (cmdq->client != NULL && cmdq->client->session == NULL)
@@ -158,7 +159,6 @@ cmd_split_window_exec(struct cmd *self, struct cmd_q *cmdq)
 	if (window_pane_spawn(new_wp, argc, argv, path, shell, cwd, &env,
 	    s->tio, &cause) != 0)
 		goto error;
-	layout_assign_pane(lc, new_wp);
 
 	server_redraw_window(w);
 
@@ -193,8 +193,10 @@ cmd_split_window_exec(struct cmd *self, struct cmd_q *cmdq)
 
 error:
 	environ_free(&env);
-	if (new_wp != NULL)
+	if (new_wp != NULL) {
+		layout_close_pane(new_wp);
 		window_remove_pane(w, new_wp);
+	}
 	cmdq_error(cmdq, "create pane failed: %s", cause);
 	free(cause);
 	if (fd != -1)
