@@ -63,13 +63,15 @@ cmd_swap_pane_exec(struct cmd *self, struct cmd_q *cmdq)
 			if (src_wp == NULL)
 				src_wp = TAILQ_LAST(&dst_w->panes, window_panes);
 		} else {
-			src_wl = cmd_find_pane(cmdq, NULL, NULL, &src_wp);
+			src_wl = cmd_find_pane_marked(cmdq, NULL, NULL,
+			    &src_wp);
 			if (src_wl == NULL)
 				return (CMD_RETURN_ERROR);
 			src_w = src_wl->window;
 		}
 	} else {
-		src_wl = cmd_find_pane(cmdq, args_get(args, 's'), NULL, &src_wp);
+		src_wl = cmd_find_pane_marked(cmdq, args_get(args, 's'), NULL,
+		    &src_wp);
 		if (src_wl == NULL)
 			return (CMD_RETURN_ERROR);
 		src_w = src_wl->window;
