@@ -40,8 +40,12 @@ openat(int fd, const char *path, int flags, ...)
 		dotfd = open(".", O_RDONLY);
 		if (dotfd == -1)
 			return (-1);
-		if (fchdir(fd) != 0)
+		if (fchdir(fd) != 0) {
+			saved_errno = errno;
+			close(dotfd);
+			errno = saved_errno;
 			return (-1);
+		}
 	}
 
 	retval = open(path, flags, mode);
