@@ -1,4 +1,4 @@
-/* $Id$ */
+/* $OpenBSD$ */
 
 /*
  * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -20,6 +20,7 @@
 
 #include <stdlib.h>
 #include <string.h>
+#include <time.h>
 
 #include "tmux.h"
 
@@ -43,18 +44,69 @@
  *   pretty stupid but not supporting it is more trouble than it is worth.
  *
  * - Special handling for ESC inside a DCS to allow arbitrary byte sequences to
- *   be passed to the underlying teminal(s).
+ *   be passed to the underlying terminals.
  */
 
+/* Input parser cell. */
+struct input_cell {
+	struct grid_cell	cell;
+	int			set;
+	int			g0set;	/* 1 if ACS */
+	int			g1set;	/* 1 if ACS */
+};
+
+/* Input parser context. */
+struct input_ctx {
+	struct window_pane     *wp;
+	struct screen_write_ctx ctx;
+
+	struct input_cell	cell;
+
+	struct input_cell	old_cell;
+	u_int 			old_cx;
+	u_int			old_cy;
+
+	u_char			interm_buf[4];
+	size_t			interm_len;
+
+	u_char			param_buf[64];
+	size_t			param_len;
+
+#define INPUT_BUF_START 32
+#define INPUT_BUF_LIMIT 1048576
+	u_char		       *input_buf;
+	size_t			input_len;
+	size_t			input_space;
+
+	int			param_list[24];	/* -1 not present */
+	u_int			param_list_len;
+
+	struct utf8_data	utf8data;
+
+	int			ch;
+	int			flags;
+#define INPUT_DISCARD 0x1
+
+	const struct input_state *state;
+
+	/*
+	 * All input received since we were last in the ground state. Sent to
+	 * control clients on connection.
+	 */
+	struct evbuffer	 	*since_ground;
+};
+
 /* Helper functions. */
 struct input_transition;
 int	input_split(struct input_ctx *);
 int	input_get(struct input_ctx *, u_int, int, int);
 void	input_reply(struct input_ctx *, const char *, ...);
 void	input_set_state(struct window_pane *, const struct input_transition *);
+void	input_reset_cell(struct input_ctx *);
 
 /* Transition entry/exit handlers. */
 void	input_clear(struct input_ctx *);
+void	input_ground(struct input_ctx *);
 void	input_enter_osc(struct input_ctx *);
 void	input_exit_osc(struct input_ctx *);
 void	input_enter_apc(struct input_ctx *);
@@ -75,6 +127,8 @@ void	input_csi_dispatch_rm_private(struct input_ctx *);
 void	input_csi_dispatch_sm(struct input_ctx *);
 void	input_csi_dispatch_sm_private(struct input_ctx *);
 void	input_csi_dispatch_winops(struct input_ctx *);
+void	input_csi_dispatch_sgr_256(struct input_ctx *, int, u_int *);
+void	input_csi_dispatch_sgr_rgb(struct input_ctx *, int, u_int *);
 void	input_csi_dispatch_sgr(struct input_ctx *);
 int	input_dcs_dispatch(struct input_ctx *);
 int	input_utf8_open(struct input_ctx *);
@@ -103,19 +157,23 @@ enum input_esc_type {
 	INPUT_ESC_NEL,
 	INPUT_ESC_RI,
 	INPUT_ESC_RIS,
-	INPUT_ESC_SCSOFF_G0,
-	INPUT_ESC_SCSON_G0,
+	INPUT_ESC_SCSG0_OFF,
+	INPUT_ESC_SCSG0_ON,
+	INPUT_ESC_SCSG1_OFF,
+	INPUT_ESC_SCSG1_ON,
 };
 
 /* Escape command table. */
 const struct input_table_entry input_esc_table[] = {
-	{ '0', "(", INPUT_ESC_SCSOFF_G0 },
+	{ '0', "(", INPUT_ESC_SCSG0_ON },
+	{ '0', ")", INPUT_ESC_SCSG1_ON },
 	{ '7', "",  INPUT_ESC_DECSC },
 	{ '8', "",  INPUT_ESC_DECRC },
 	{ '8', "#", INPUT_ESC_DECALN },
 	{ '=', "",  INPUT_ESC_DECKPAM },
 	{ '>', "",  INPUT_ESC_DECKPNM },
-	{ 'B', "(", INPUT_ESC_SCSON_G0 },
+	{ 'B', "(", INPUT_ESC_SCSG0_OFF },
+	{ 'B', ")", INPUT_ESC_SCSG1_OFF },
 	{ 'D', "",  INPUT_ESC_IND },
 	{ 'E', "",  INPUT_ESC_NEL },
 	{ 'H', "",  INPUT_ESC_HTS },
@@ -242,7 +300,7 @@ const struct input_transition input_state_utf8_one_table[];
 /* ground state definition. */
 const struct input_state input_state_ground = {
 	"ground",
-	NULL, NULL,
+	input_ground, NULL,
 	input_state_ground_table
 };
 
@@ -683,17 +741,28 @@ input_table_compare(const void *key, const void *value)
 	return (strcmp(ictx->interm_buf, entry->interm));
 }
 
+/* Reset cell state to default. */
+void
+input_reset_cell(struct input_ctx *ictx)
+{
+	memcpy(&ictx->cell.cell, &grid_default_cell, sizeof ictx->cell.cell);
+	ictx->cell.set = 0;
+	ictx->cell.g0set = ictx->cell.g1set = 0;
+
+	memcpy(&ictx->old_cell, &ictx->cell, sizeof ictx->old_cell);
+	ictx->old_cx = 0;
+	ictx->old_cy = 0;
+}
+
 /* Initialise input parser. */
 void
 input_init(struct window_pane *wp)
 {
-	struct input_ctx	*ictx = &wp->ictx;
+	struct input_ctx	*ictx;
 
-	memcpy(&ictx->cell, &grid_default_cell, sizeof ictx->cell);
+	ictx = wp->ictx = xcalloc(1, sizeof *ictx);
 
-	memcpy(&ictx->old_cell, &grid_default_cell, sizeof ictx->old_cell);
-	ictx->old_cx = 0;
-	ictx->old_cy = 0;
+	input_reset_cell(ictx);
 
 	*ictx->interm_buf = '\0';
 	ictx->interm_len = 0;
@@ -701,6 +770,12 @@ input_init(struct window_pane *wp)
 	*ictx->param_buf = '\0';
 	ictx->param_len = 0;
 
+	ictx->input_space = INPUT_BUF_START;
+	ictx->input_buf = xmalloc(INPUT_BUF_START);
+
+	*ictx->input_buf = '\0';
+	ictx->input_len = 0;
+
 	ictx->state = &input_state_ground;
 	ictx->flags = 0;
 
@@ -711,23 +786,46 @@ input_init(struct window_pane *wp)
 void
 input_free(struct window_pane *wp)
 {
-	if (wp != NULL)
-		evbuffer_free(wp->ictx.since_ground);
+	struct input_ctx	*ictx = wp->ictx;
+
+	free(ictx->input_buf);
+	evbuffer_free(ictx->since_ground);
+
+	free (ictx);
+	wp->ictx = NULL;
+}
+
+/* Reset input state and clear screen. */
+void
+input_reset(struct window_pane *wp)
+{
+	struct input_ctx	*ictx = wp->ictx;
+
+	input_reset_cell(ictx);
+
+	if (wp->mode == NULL)
+		screen_write_start(&ictx->ctx, wp, &wp->base);
+	else
+		screen_write_start(&ictx->ctx, NULL, &wp->base);
+	screen_write_reset(&ictx->ctx);
+	screen_write_stop(&ictx->ctx);
+}
+
+/* Return pending data. */
+struct evbuffer *
+input_pending(struct window_pane *wp)
+{
+	return (wp->ictx->since_ground);
 }
 
 /* Change input state. */
 void
 input_set_state(struct window_pane *wp, const struct input_transition *itr)
 {
-	struct input_ctx	*ictx = &wp->ictx;
-	struct evbuffer		*ground_evb = ictx->since_ground;
+	struct input_ctx	*ictx = wp->ictx;
 
 	if (ictx->state->exit != NULL)
 		ictx->state->exit(ictx);
-
-	if (itr->state == &input_state_ground)
-		evbuffer_drain(ground_evb, EVBUFFER_LENGTH(ground_evb));
-
 	ictx->state = itr->state;
 	if (ictx->state->enter != NULL)
 		ictx->state->enter(ictx);
@@ -737,7 +835,7 @@ input_set_state(struct window_pane *wp, const struct input_transition *itr)
 void
 input_parse(struct window_pane *wp)
 {
-	struct input_ctx		*ictx = &wp->ictx;
+	struct input_ctx		*ictx = wp->ictx;
 	const struct input_transition	*itr;
 	struct evbuffer			*evb = wp->event->input;
 	u_char				*buf;
@@ -746,8 +844,8 @@ input_parse(struct window_pane *wp)
 	if (EVBUFFER_LENGTH(evb) == 0)
 		return;
 
-	wp->window->flags |= WINDOW_ACTIVITY;
-	wp->window->flags &= ~WINDOW_SILENCE;
+	window_update_activity(wp->window);
+	wp->flags |= PANE_CHANGED;
 
 	/*
 	 * Open the screen. Use NULL wp if there is a mode set as don't want to
@@ -764,10 +862,12 @@ input_parse(struct window_pane *wp)
 	notify_input(wp, evb);
 	off = 0;
 
+	log_debug("%s: %%%u %s, %zu bytes: %.*s", __func__, wp->id,
+	    ictx->state->name, len, (int)len, buf);
+
 	/* Parse the input. */
 	while (off < len) {
 		ictx->ch = buf[off++];
-		log_debug("%s: '%c' %s", __func__, ictx->ch, ictx->state->name);
 
 		/* Find the transition. */
 		itr = ictx->state->transitions;
@@ -778,7 +878,7 @@ input_parse(struct window_pane *wp)
 		}
 		if (itr->first == -1 || itr->last == -1) {
 			/* No transition? Eh? */
-			fatalx("No transition from state!");
+			fatalx("no transition from state");
 		}
 
 		/*
@@ -882,12 +982,34 @@ input_clear(struct input_ctx *ictx)
 	ictx->flags &= ~INPUT_DISCARD;
 }
 
+/* Reset for ground state. */
+void
+input_ground(struct input_ctx *ictx)
+{
+	evbuffer_drain(ictx->since_ground, EVBUFFER_LENGTH(ictx->since_ground));
+
+	if (ictx->input_space > INPUT_BUF_START) {
+		ictx->input_space = INPUT_BUF_START;
+		ictx->input_buf = xrealloc(ictx->input_buf, INPUT_BUF_START);
+	}
+}
+
 /* Output this character to the screen. */
 int
 input_print(struct input_ctx *ictx)
 {
-	grid_cell_one(&ictx->cell, ictx->ch);
-	screen_write_cell(&ictx->ctx, &ictx->cell);
+	int	set;
+
+	set = ictx->cell.set == 0 ? ictx->cell.g0set : ictx->cell.g1set;
+	if (set == 1)
+		ictx->cell.cell.attr |= GRID_ATTR_CHARSET;
+	else
+		ictx->cell.cell.attr &= ~GRID_ATTR_CHARSET;
+
+	grid_cell_one(&ictx->cell.cell, ictx->ch);
+	screen_write_cell(&ictx->ctx, &ictx->cell.cell);
+
+	ictx->cell.cell.attr &= ~GRID_ATTR_CHARSET;
 
 	return (0);
 }
@@ -924,12 +1046,20 @@ input_parameter(struct input_ctx *ictx)
 int
 input_input(struct input_ctx *ictx)
 {
-	if (ictx->input_len == (sizeof ictx->input_buf) - 1)
-		ictx->flags |= INPUT_DISCARD;
-	else {
-		ictx->input_buf[ictx->input_len++] = ictx->ch;
-		ictx->input_buf[ictx->input_len] = '\0';
+	size_t available;
+
+	available = ictx->input_space;
+	while (ictx->input_len + 1 >= available) {
+		available *= 2;
+		if (available > INPUT_BUF_LIMIT) {
+			ictx->flags |= INPUT_DISCARD;
+			return (0);
+		}
+		ictx->input_buf = xrealloc(ictx->input_buf, available);
+		ictx->input_space = available;
 	}
+	ictx->input_buf[ictx->input_len++] = ictx->ch;
+	ictx->input_buf[ictx->input_len] = '\0';
 
 	return (0);
 }
@@ -941,19 +1071,18 @@ input_c0_dispatch(struct input_ctx *ictx)
 	struct screen_write_ctx	*sctx = &ictx->ctx;
 	struct window_pane	*wp = ictx->wp;
 	struct screen		*s = sctx->s;
-	u_int			 trigger;
 
-	log_debug("%s: '%c", __func__, ictx->ch);
+	log_debug("%s: '%c'", __func__, ictx->ch);
 
 	switch (ictx->ch) {
 	case '\000':	/* NUL */
 		break;
 	case '\007':	/* BEL */
-		wp->window->flags |= WINDOW_BELL;
+		alerts_queue(wp->window, WINDOW_BELL);
 		break;
 	case '\010':	/* BS */
 		screen_write_backspace(sctx);
-		goto count_c0;
+		break;
 	case '\011':	/* HT */
 		/* Don't tab beyond the end of the line. */
 		if (s->cx >= screen_size_x(s) - 1)
@@ -970,30 +1099,21 @@ input_c0_dispatch(struct input_ctx *ictx)
 	case '\013':	/* VT */
 	case '\014':	/* FF */
 		screen_write_linefeed(sctx, 0);
-		goto count_c0;
+		break;
 	case '\015':	/* CR */
 		screen_write_carriagereturn(sctx);
-		goto count_c0;
+		break;
 	case '\016':	/* SO */
-		ictx->cell.attr |= GRID_ATTR_CHARSET;
+		ictx->cell.set = 1;
 		break;
 	case '\017':	/* SI */
-		ictx->cell.attr &= ~GRID_ATTR_CHARSET;
+		ictx->cell.set = 0;
 		break;
 	default:
 		log_debug("%s: unknown '%c'", __func__, ictx->ch);
 		break;
 	}
 
-	return (0);
-
-count_c0:
-	trigger = options_get_number(&wp->window->options, "c0-change-trigger");
-	if (++wp->changes == trigger) {
-		wp->flags |= PANE_DROP;
-		window_pane_timer_start(wp);
-	}
-
 	return (0);
 }
 
@@ -1018,11 +1138,7 @@ input_esc_dispatch(struct input_ctx *ictx)
 
 	switch (entry->type) {
 	case INPUT_ESC_RIS:
-		memcpy(&ictx->cell, &grid_default_cell, sizeof ictx->cell);
-		memcpy(&ictx->old_cell, &ictx->cell, sizeof ictx->old_cell);
-		ictx->old_cx = 0;
-		ictx->old_cy = 0;
-
+		input_reset_cell(ictx);
 		screen_write_reset(sctx);
 		break;
 	case INPUT_ESC_IND:
@@ -1057,16 +1173,17 @@ input_esc_dispatch(struct input_ctx *ictx)
 	case INPUT_ESC_DECALN:
 		screen_write_alignmenttest(sctx);
 		break;
-	case INPUT_ESC_SCSON_G0:
-		/*
-		 * Not really supported, but fake it up enough for those that
-		 * use it to switch character sets (by redefining G0 to
-		 * graphics set, rather than switching to G1).
-		 */
-		ictx->cell.attr &= ~GRID_ATTR_CHARSET;
+	case INPUT_ESC_SCSG0_ON:
+		ictx->cell.g0set = 1;
 		break;
-	case INPUT_ESC_SCSOFF_G0:
-		ictx->cell.attr |= GRID_ATTR_CHARSET;
+	case INPUT_ESC_SCSG0_OFF:
+		ictx->cell.g0set = 0;
+		break;
+	case INPUT_ESC_SCSG1_ON:
+		ictx->cell.g1set = 1;
+		break;
+	case INPUT_ESC_SCSG1_OFF:
+		ictx->cell.g1set = 0;
 		break;
 	}
 
@@ -1081,6 +1198,7 @@ input_csi_dispatch(struct input_ctx *ictx)
 	struct screen		       *s = sctx->s;
 	struct input_table_entry       *entry;
 	int				n, m;
+	u_int				cx;
 
 	if (ictx->flags & INPUT_DISCARD)
 		return (0);
@@ -1099,12 +1217,16 @@ input_csi_dispatch(struct input_ctx *ictx)
 	switch (entry->type) {
 	case INPUT_CSI_CBT:
 		/* Find the previous tab point, n times. */
+		cx = s->cx;
+		if (cx > screen_size_x(s) - 1)
+			cx = screen_size_x(s) - 1;
 		n = input_get(ictx, 0, 1, 1);
-		while (s->cx > 0 && n-- > 0) {
+		while (cx > 0 && n-- > 0) {
 			do
-				s->cx--;
-			while (s->cx > 0 && !bit_test(s->tabs, s->cx));
+				cx--;
+			while (cx > 0 && !bit_test(s->tabs, cx));
 		}
+		s->cx = cx;
 		break;
 	case INPUT_CSI_CUB:
 		screen_write_cursorleft(sctx, input_get(ictx, 0, 1, 1));
@@ -1147,7 +1269,7 @@ input_csi_dispatch(struct input_ctx *ictx)
 	case INPUT_CSI_DA_TWO:
 		switch (input_get(ictx, 0, 0, 0)) {
 		case 0:
-			input_reply(ictx, "\033[>0;95;0c");
+			input_reply(ictx, "\033[>84;0;0c");
 			break;
 		default:
 			log_debug("%s: unknown '%c'", __func__, ictx->ch);
@@ -1296,6 +1418,9 @@ input_csi_dispatch_rm(struct input_ctx *ictx)
 		case 4:		/* IRM */
 			screen_write_mode_clear(&ictx->ctx, MODE_INSERT);
 			break;
+		case 34:
+			screen_write_mode_set(&ictx->ctx, MODE_BLINKING);
+			break;
 		default:
 			log_debug("%s: unknown '%c'", __func__, ictx->ch);
 			break;
@@ -1307,7 +1432,8 @@ input_csi_dispatch_rm(struct input_ctx *ictx)
 void
 input_csi_dispatch_rm_private(struct input_ctx *ictx)
 {
-	u_int	i;
+	struct window_pane	*wp = ictx->wp;
+	u_int			 i;
 
 	for (i = 0; i < ictx->param_list_len; i++) {
 		switch (input_get(ictx, i, 0, -1)) {
@@ -1321,13 +1447,15 @@ input_csi_dispatch_rm_private(struct input_ctx *ictx)
 		case 7:		/* DECAWM */
 			screen_write_mode_clear(&ictx->ctx, MODE_WRAP);
 			break;
+		case 12:
+			screen_write_mode_clear(&ictx->ctx, MODE_BLINKING);
+			break;
 		case 25:	/* TCEM */
 			screen_write_mode_clear(&ictx->ctx, MODE_CURSOR);
 			break;
 		case 1000:
 		case 1001:
 		case 1002:
-		case 1003:
 			screen_write_mode_clear(&ictx->ctx, ALL_MOUSE_MODES);
 			break;
 		case 1004:
@@ -1341,10 +1469,10 @@ input_csi_dispatch_rm_private(struct input_ctx *ictx)
 			break;
 		case 47:
 		case 1047:
-			window_pane_alternate_off(ictx->wp, &ictx->cell, 0);
+			window_pane_alternate_off(wp, &ictx->cell.cell, 0);
 			break;
 		case 1049:
-			window_pane_alternate_off(ictx->wp, &ictx->cell, 1);
+			window_pane_alternate_off(wp, &ictx->cell.cell, 1);
 			break;
 		case 2004:
 			screen_write_mode_clear(&ictx->ctx, MODE_BRACKETPASTE);
@@ -1367,6 +1495,9 @@ input_csi_dispatch_sm(struct input_ctx *ictx)
 		case 4:		/* IRM */
 			screen_write_mode_set(&ictx->ctx, MODE_INSERT);
 			break;
+		case 34:
+			screen_write_mode_clear(&ictx->ctx, MODE_BLINKING);
+			break;
 		default:
 			log_debug("%s: unknown '%c'", __func__, ictx->ch);
 			break;
@@ -1378,7 +1509,8 @@ input_csi_dispatch_sm(struct input_ctx *ictx)
 void
 input_csi_dispatch_sm_private(struct input_ctx *ictx)
 {
-	u_int	i;
+	struct window_pane	*wp = ictx->wp;
+	u_int			 i;
 
 	for (i = 0; i < ictx->param_list_len; i++) {
 		switch (input_get(ictx, i, 0, -1)) {
@@ -1392,6 +1524,9 @@ input_csi_dispatch_sm_private(struct input_ctx *ictx)
 		case 7:		/* DECAWM */
 			screen_write_mode_set(&ictx->ctx, MODE_WRAP);
 			break;
+		case 12:
+			screen_write_mode_set(&ictx->ctx, MODE_BLINKING);
+			break;
 		case 25:	/* TCEM */
 			screen_write_mode_set(&ictx->ctx, MODE_CURSOR);
 			break;
@@ -1403,15 +1538,11 @@ input_csi_dispatch_sm_private(struct input_ctx *ictx)
 			screen_write_mode_clear(&ictx->ctx, ALL_MOUSE_MODES);
 			screen_write_mode_set(&ictx->ctx, MODE_MOUSE_BUTTON);
 			break;
-		case 1003:
-			screen_write_mode_clear(&ictx->ctx, ALL_MOUSE_MODES);
-			screen_write_mode_set(&ictx->ctx, MODE_MOUSE_ANY);
-			break;
 		case 1004:
 			if (ictx->ctx.s->mode & MODE_FOCUSON)
 				break;
 			screen_write_mode_set(&ictx->ctx, MODE_FOCUSON);
-			ictx->wp->flags |= PANE_FOCUSPUSH; /* force update */
+			wp->flags |= PANE_FOCUSPUSH; /* force update */
 			break;
 		case 1005:
 			screen_write_mode_set(&ictx->ctx, MODE_MOUSE_UTF8);
@@ -1421,10 +1552,10 @@ input_csi_dispatch_sm_private(struct input_ctx *ictx)
 			break;
 		case 47:
 		case 1047:
-			window_pane_alternate_on(ictx->wp, &ictx->cell, 0);
+			window_pane_alternate_on(wp, &ictx->cell.cell, 0);
 			break;
 		case 1049:
-			window_pane_alternate_on(ictx->wp, &ictx->cell, 1);
+			window_pane_alternate_on(wp, &ictx->cell.cell, 1);
 			break;
 		case 2004:
 			screen_write_mode_set(&ictx->ctx, MODE_BRACKETPASTE);
@@ -1475,7 +1606,7 @@ input_csi_dispatch_winops(struct input_ctx *ictx)
 				return;
 			break;
 		case 18:
-			input_reply(ictx, "\033[8;%u;%u", wp->sy, wp->sx);
+			input_reply(ictx, "\033[8;%u;%ut", wp->sy, wp->sx);
 			break;
 		default:
 			log_debug("%s: unknown '%c'", __func__, ictx->ch);
@@ -1485,19 +1616,74 @@ input_csi_dispatch_winops(struct input_ctx *ictx)
 	}
 }
 
+/* Handle CSI SGR for 256 colours. */
+void
+input_csi_dispatch_sgr_256(struct input_ctx *ictx, int fgbg, u_int *i)
+{
+	struct grid_cell	*gc = &ictx->cell.cell;
+	int			 c;
+
+	(*i)++;
+	c = input_get(ictx, *i, 0, -1);
+	if (c == -1) {
+		if (fgbg == 38) {
+			gc->flags &= ~GRID_FLAG_FG256;
+			gc->fg = 8;
+		} else if (fgbg == 48) {
+			gc->flags &= ~GRID_FLAG_BG256;
+			gc->bg = 8;
+		}
+	} else {
+		if (fgbg == 38) {
+			gc->flags |= GRID_FLAG_FG256;
+			gc->fg = c;
+		} else if (fgbg == 48) {
+			gc->flags |= GRID_FLAG_BG256;
+			gc->bg = c;
+		}
+	}
+}
+
+/* Handle CSI SGR for RGB colours. */
+void
+input_csi_dispatch_sgr_rgb(struct input_ctx *ictx, int fgbg, u_int *i)
+{
+	struct grid_cell	*gc = &ictx->cell.cell;
+	int			 c, r, g, b;
+
+	(*i)++;
+	r = input_get(ictx, *i, 0, -1);
+	if (r == -1 || r > 255)
+		return;
+	(*i)++;
+	g = input_get(ictx, *i, 0, -1);
+	if (g == -1 || g > 255)
+		return;
+	(*i)++;
+	b = input_get(ictx, *i, 0, -1);
+	if (b == -1 || b > 255)
+		return;
+
+	c = colour_find_rgb(r, g, b);
+	if (fgbg == 38) {
+		gc->flags |= GRID_FLAG_FG256;
+		gc->fg = c;
+	} else if (fgbg == 48) {
+		gc->flags |= GRID_FLAG_BG256;
+		gc->bg = c;
+	}
+}
+
 /* Handle CSI SGR. */
 void
 input_csi_dispatch_sgr(struct input_ctx *ictx)
 {
-	struct grid_cell	*gc = &ictx->cell;
+	struct grid_cell	*gc = &ictx->cell.cell;
 	u_int			 i;
-	int			 n, m;
-	u_char			 attr;
+	int			 n;
 
 	if (ictx->param_list_len == 0) {
-		attr = gc->attr;
 		memcpy(gc, &grid_default_cell, sizeof *gc);
-		gc->attr |= (attr & GRID_ATTR_CHARSET);
 		return;
 	}
 
@@ -1506,28 +1692,13 @@ input_csi_dispatch_sgr(struct input_ctx *ictx)
 
 		if (n == 38 || n == 48) {
 			i++;
-			if (input_get(ictx, i, 0, -1) != 5)
-				continue;
-
-			i++;
-			m = input_get(ictx, i, 0, -1);
-			if (m == -1) {
-				if (n == 38) {
-					gc->flags &= ~GRID_FLAG_FG256;
-					gc->fg = 8;
-				} else if (n == 48) {
-					gc->flags &= ~GRID_FLAG_BG256;
-					gc->bg = 8;
-				}
-
-			} else {
-				if (n == 38) {
-					gc->flags |= GRID_FLAG_FG256;
-					gc->fg = m;
-				} else if (n == 48) {
-					gc->flags |= GRID_FLAG_BG256;
-					gc->bg = m;
-				}
+			switch (input_get(ictx, i, 0, -1)) {
+			case 2:
+				input_csi_dispatch_sgr_rgb(ictx, n, &i);
+				break;
+			case 5:
+				input_csi_dispatch_sgr_256(ictx, n, &i);
+				break;
 			}
 			continue;
 		}
@@ -1535,9 +1706,7 @@ input_csi_dispatch_sgr(struct input_ctx *ictx)
 		switch (n) {
 		case 0:
 		case 10:
-			attr = gc->attr;
 			memcpy(gc, &grid_default_cell, sizeof *gc);
-			gc->attr |= (attr & GRID_ATTR_CHARSET);
 			break;
 		case 1:
 			gc->attr |= GRID_ATTR_BRIGHT;
@@ -1666,8 +1835,8 @@ input_enter_osc(struct input_ctx *ictx)
 void
 input_exit_osc(struct input_ctx *ictx)
 {
-	u_char *p = ictx->input_buf;
-	int	option;
+	u_char	*p = ictx->input_buf;
+	u_int	 option;
 
 	if (ictx->flags & INPUT_DISCARD)
 		return;
@@ -1781,8 +1950,8 @@ input_utf8_close(struct input_ctx *ictx)
 
 	utf8_append(&ictx->utf8data, ictx->ch);
 
-	grid_cell_set(&ictx->cell, &ictx->utf8data);
-	screen_write_cell(&ictx->ctx, &ictx->cell);
+	grid_cell_set(&ictx->cell.cell, &ictx->utf8data);
+	screen_write_cell(&ictx->ctx, &ictx->cell.cell);
 
 	return (0);
 }
