@@ -30,7 +30,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <syslog.h>
 #include <termios.h>
 #include <time.h>
 #include <unistd.h>
@@ -41,28 +40,87 @@
  * Main server functions.
  */
 
-/* Client list. */
 struct clients	 clients;
-struct clients	 dead_clients;
 
 int		 server_fd;
 int		 server_shutdown;
 struct event	 server_ev_accept;
-struct event	 server_ev_second;
 
-int		 server_create_socket(void);
-void		 server_loop(void);
-int		 server_should_shutdown(void);
-void		 server_send_shutdown(void);
-void		 server_clean_dead(void);
-void		 server_accept_callback(int, short, void *);
-void		 server_signal_callback(int, short, void *);
-void		 server_child_signal(void);
-void		 server_child_exited(pid_t, int);
-void		 server_child_stopped(pid_t, int);
-void		 server_second_callback(int, short, void *);
-void		 server_lock_server(void);
-void		 server_lock_sessions(void);
+struct session		*marked_session;
+struct winlink		*marked_winlink;
+struct window		*marked_window;
+struct window_pane	*marked_window_pane;
+struct layout_cell	*marked_layout_cell;
+
+int	server_create_socket(void);
+void	server_loop(void);
+int	server_should_shutdown(void);
+void	server_send_shutdown(void);
+void	server_accept_callback(int, short, void *);
+void	server_signal_callback(int, short, void *);
+void	server_child_signal(void);
+void	server_child_exited(pid_t, int);
+void	server_child_stopped(pid_t, int);
+
+/* Set marked pane. */
+void
+server_set_marked(struct session *s, struct winlink *wl, struct window_pane *wp)
+{
+	marked_session = s;
+	marked_winlink = wl;
+	marked_window = wl->window;
+	marked_window_pane = wp;
+	marked_layout_cell = wp->layout_cell;
+}
+
+/* Clear marked pane. */
+void
+server_clear_marked(void)
+{
+	marked_session = NULL;
+	marked_winlink = NULL;
+	marked_window = NULL;
+	marked_window_pane = NULL;
+	marked_layout_cell = NULL;
+}
+
+/* Is this the marked pane? */
+int
+server_is_marked(struct session *s, struct winlink *wl, struct window_pane *wp)
+{
+	if (s == NULL || wl == NULL || wp == NULL)
+		return (0);
+	if (marked_session != s || marked_winlink != wl)
+		return (0);
+	if (marked_window_pane != wp)
+		return (0);
+	return (server_check_marked());
+}
+
+/* Check if the marked pane is still valid. */
+int
+server_check_marked(void)
+{
+	struct winlink	*wl;
+
+	if (marked_window_pane == NULL)
+		return (0);
+	if (marked_layout_cell != marked_window_pane->layout_cell)
+		return (0);
+
+	if (!session_alive(marked_session))
+		return (0);
+	RB_FOREACH(wl, winlinks, &marked_session->windows) {
+		if (wl->window == marked_window && wl == marked_winlink)
+			break;
+	}
+	if (wl == NULL)
+		return (0);
+
+	if (!window_has_pane(marked_window, marked_window_pane))
+		return (0);
+	return (window_pane_visible(marked_window_pane));
+}
 
 /* Create server socket. */
 int
@@ -78,34 +136,30 @@ server_create_socket(void)
 	size = strlcpy(sa.sun_path, socket_path, sizeof sa.sun_path);
 	if (size >= sizeof sa.sun_path) {
 		errno = ENAMETOOLONG;
-		fatal("socket failed");
+		return (-1);
 	}
 	unlink(sa.sun_path);
 
 	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
-		fatal("socket failed");
+		return (-1);
 
 	mask = umask(S_IXUSR|S_IXGRP|S_IRWXO);
-	if (bind(fd, (struct sockaddr *) &sa, SUN_LEN(&sa)) == -1)
-		fatal("bind failed");
+	if (bind(fd, (struct sockaddr *) &sa, sizeof(sa)) == -1)
+		return (-1);
 	umask(mask);
 
 	if (listen(fd, 16) == -1)
-		fatal("listen failed");
+		return (-1);
 	setblocking(fd, 0);
 
-	server_update_socket();
-
 	return (fd);
 }
 
 /* Fork new server. */
 int
-server_start(int lockfd, char *lockfile)
+server_start(struct event_base *base, int lockfd, char *lockfile)
 {
-	int	 	 pair[2];
-	struct timeval	 tv;
-	char		*cause;
+	int	pair[2];
 
 	/* The first client is special and gets a socketpair; create it. */
 	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pair) != 0)
@@ -131,19 +185,17 @@ server_start(int lockfd, char *lockfile)
 		fatal("daemon failed");
 
 	/* event_init() was called in our parent, need to reinit. */
-	if (event_reinit(ev_base) != 0)
-		fatal("event_reinit failed");
 	clear_signals(0);
+	if (event_reinit(base) != 0)
+		fatal("event_reinit failed");
 
 	logfile("server");
 	log_debug("server started, pid %ld", (long) getpid());
 
-	ARRAY_INIT(&windows);
+	RB_INIT(&windows);
 	RB_INIT(&all_window_panes);
-	ARRAY_INIT(&clients);
-	ARRAY_INIT(&dead_clients);
+	TAILQ_INIT(&clients);
 	RB_INIT(&sessions);
-	RB_INIT(&dead_sessions);
 	TAILQ_INIT(&session_groups);
 	mode_key_init_trees();
 	key_bindings_init();
@@ -156,40 +208,24 @@ server_start(int lockfd, char *lockfile)
 #endif
 
 	server_fd = server_create_socket();
+	if (server_fd == -1)
+		fatal("couldn't create socket");
+	server_update_socket();
 	server_client_create(pair[1]);
 
 	unlink(lockfile);
 	free(lockfile);
 	close(lockfd);
 
-	cfg_cmd_q = cmdq_new(NULL);
-	cfg_cmd_q->emptyfn = cfg_default_done;
-	cfg_finished = 0;
-	cfg_references = 1;
-	cfg_client = ARRAY_FIRST(&clients);
-	if (cfg_client != NULL)
-		cfg_client->references++;
+	start_cfg();
 
-	if (access(TMUX_CONF, R_OK) == 0) {
-		if (load_cfg(TMUX_CONF, cfg_cmd_q, &cause) == -1)
-			cfg_add_cause("%s: %s", TMUX_CONF, cause);
-	} else if (errno != ENOENT)
-		cfg_add_cause("%s: %s", TMUX_CONF, strerror(errno));
-	if (cfg_file != NULL) {
-		if (load_cfg(cfg_file, cfg_cmd_q, &cause) == -1)
-			cfg_add_cause("%s: %s", cfg_file, cause);
-	}
-	cmdq_continue(cfg_cmd_q);
+	status_prompt_load_history();
 
 	server_add_accept(0);
 
-	memset(&tv, 0, sizeof tv);
-	tv.tv_sec = 1;
-	evtimer_set(&server_ev_second, server_second_callback, NULL);
-	evtimer_add(&server_ev_second, &tv);
-
 	set_signals(server_signal_callback);
 	server_loop();
+	status_prompt_save_history();
 	exit(0);
 }
 
@@ -198,12 +234,11 @@ void
 server_loop(void)
 {
 	while (!server_should_shutdown()) {
+		log_debug("event dispatch enter");
 		event_loop(EVLOOP_ONCE);
+		log_debug("event dispatch exit");
 
-		server_window_loop();
 		server_client_loop();
-
-		server_clean_dead();
 	}
 }
 
@@ -212,16 +247,14 @@ int
 server_should_shutdown(void)
 {
 	struct client	*c;
-	u_int		 i;
 
 	if (!options_get_number(&global_options, "exit-unattached")) {
 		if (!RB_EMPTY(&sessions))
 			return (0);
 	}
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c != NULL && c->session != NULL)
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->session != NULL)
 			return (0);
 	}
 
@@ -230,10 +263,8 @@ server_should_shutdown(void)
 	 * clients but don't actually exit until they've gone.
 	 */
 	cmd_wait_for_flush();
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		if (ARRAY_ITEM(&clients, i) != NULL)
-			return (0);
-	}
+	if (!TAILQ_EMPTY(&clients))
+		return (0);
 
 	return (1);
 }
@@ -242,57 +273,21 @@ server_should_shutdown(void)
 void
 server_send_shutdown(void)
 {
-	struct client	*c;
-	struct session	*s, *next_s;
-	u_int		 i;
+	struct client	*c, *c1;
+	struct session	*s, *s1;
 
 	cmd_wait_for_flush();
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c != NULL) {
-			if (c->flags & (CLIENT_BAD|CLIENT_SUSPENDED))
-				server_client_lost(c);
-			else
-				server_write_client(c, MSG_SHUTDOWN, NULL, 0);
-			c->session = NULL;
-		}
+	TAILQ_FOREACH_SAFE(c, &clients, entry, c1) {
+		if (c->flags & (CLIENT_BAD|CLIENT_SUSPENDED))
+			server_client_lost(c);
+		else
+			server_write_client(c, MSG_SHUTDOWN, NULL, 0);
+		c->session = NULL;
 	}
 
-	s = RB_MIN(sessions, &sessions);
-	while (s != NULL) {
-		next_s = RB_NEXT(sessions, &sessions, s);
+	RB_FOREACH_SAFE(s, sessions, &sessions, s1)
 		session_destroy(s);
-		s = next_s;
-	}
-}
-
-/* Free dead, unreferenced clients and sessions. */
-void
-server_clean_dead(void)
-{
-	struct session	*s, *next_s;
-	struct client	*c;
-	u_int		 i;
-
-	s = RB_MIN(sessions, &dead_sessions);
-	while (s != NULL) {
-		next_s = RB_NEXT(sessions, &dead_sessions, s);
-		if (s->references == 0) {
-			RB_REMOVE(sessions, &dead_sessions, s);
-			free(s->name);
-			free(s);
-		}
-		s = next_s;
-	}
-
-	for (i = 0; i < ARRAY_LENGTH(&dead_clients); i++) {
-		c = ARRAY_ITEM(&dead_clients, i);
-		if (c == NULL || c->references != 0)
-			continue;
-		ARRAY_SET(&dead_clients, i, NULL);
-		free(c);
-	}
 }
 
 /* Update socket execute permissions based on whether sessions are attached. */
@@ -388,6 +383,8 @@ server_add_accept(int timeout)
 void
 server_signal_callback(int sig, unused short events, unused void *data)
 {
+	int	fd;
+
 	switch (sig) {
 	case SIGTERM:
 		server_shutdown = 1;
@@ -398,8 +395,12 @@ server_signal_callback(int sig, unused short events, unused void *data)
 		break;
 	case SIGUSR1:
 		event_del(&server_ev_accept);
-		close(server_fd);
-		server_fd = server_create_socket();
+		fd = server_create_socket();
+		if (fd != -1) {
+			close(server_fd);
+			server_fd = fd;
+			server_update_socket();
+		}
 		server_add_accept(0);
 		break;
 	}
@@ -432,14 +433,11 @@ server_child_signal(void)
 void
 server_child_exited(pid_t pid, int status)
 {
-	struct window		*w;
+	struct window		*w, *w1;
 	struct window_pane	*wp;
 	struct job		*job;
-	u_int		 	 i;
 
-	for (i = 0; i < ARRAY_LENGTH(&windows); i++) {
-		if ((w = ARRAY_ITEM(&windows, i)) == NULL)
-			continue;
+	RB_FOREACH_SAFE(w, windows, &windows, w1) {
 		TAILQ_FOREACH(wp, &w->panes, entry) {
 			if (wp->pid == pid) {
 				wp->status = status;
@@ -463,14 +461,11 @@ server_child_stopped(pid_t pid, int status)
 {
 	struct window		*w;
 	struct window_pane	*wp;
-	u_int			 i;
 
 	if (WSTOPSIG(status) == SIGTTIN || WSTOPSIG(status) == SIGTTOU)
 		return;
 
-	for (i = 0; i < ARRAY_LENGTH(&windows); i++) {
-		if ((w = ARRAY_ITEM(&windows, i)) == NULL)
-			continue;
+	RB_FOREACH(w, windows, &windows) {
 		TAILQ_FOREACH(wp, &w->panes, entry) {
 			if (wp->pid == pid) {
 				if (killpg(pid, SIGCONT) != 0)
@@ -479,77 +474,3 @@ server_child_stopped(pid_t pid, int status)
 		}
 	}
 }
-
-/* Handle once-per-second timer events. */
-void
-server_second_callback(unused int fd, unused short events, unused void *arg)
-{
-	struct window		*w;
-	struct window_pane	*wp;
-	struct timeval		 tv;
-	u_int		 	 i;
-
-	if (options_get_number(&global_s_options, "lock-server"))
-		server_lock_server();
-	else
-		server_lock_sessions();
-
-	for (i = 0; i < ARRAY_LENGTH(&windows); i++) {
-		w = ARRAY_ITEM(&windows, i);
-		if (w == NULL)
-			continue;
-
-		TAILQ_FOREACH(wp, &w->panes, entry) {
-			if (wp->mode != NULL && wp->mode->timer != NULL)
-				wp->mode->timer(wp);
-		}
-	}
-
-	server_client_status_timer();
-
-	evtimer_del(&server_ev_second);
-	memset(&tv, 0, sizeof tv);
-	tv.tv_sec = 1;
-	evtimer_add(&server_ev_second, &tv);
-}
-
-/* Lock the server if ALL sessions have hit the time limit. */
-void
-server_lock_server(void)
-{
-	struct session  *s;
-	int		 timeout;
-	time_t           t;
-
-	t = time(NULL);
-	RB_FOREACH(s, sessions, &sessions) {
-		if (s->flags & SESSION_UNATTACHED)
-			continue;
-		timeout = options_get_number(&s->options, "lock-after-time");
-		if (timeout <= 0 || t <= s->activity_time.tv_sec + timeout)
-			return;	/* not timed out */
-	}
-
-	server_lock();
-	recalculate_sizes();
-}
-
-/* Lock any sessions which have timed out. */
-void
-server_lock_sessions(void)
-{
-	struct session  *s;
-	int		 timeout;
-	time_t		 t;
-
-	t = time(NULL);
-	RB_FOREACH(s, sessions, &sessions) {
-		if (s->flags & SESSION_UNATTACHED)
-			continue;
-		timeout = options_get_number(&s->options, "lock-after-time");
-		if (timeout > 0 && t > s->activity_time.tv_sec + timeout) {
-			server_lock_session(s);
-			recalculate_sizes();
-		}
-	}
-}
