@@ -1,4 +1,4 @@
-/* $Id: tty-acs.c 2553 2011-07-09 09:42:33Z tcunha $ */
+/* $OpenBSD$ */
 
 /*
  * Copyright (c) 2010 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -30,38 +30,38 @@ struct tty_acs_entry {
 	const char	*string;
 };
 const struct tty_acs_entry tty_acs_table[] = {
-	{ '+', "\342\206\222" },
-	{ ',', "\342\206\220" },
-	{ '-', "\342\206\221" },
-	{ '.', "\342\206\223" },
-	{ '0', "\342\226\256" },
-	{ '`', "\342\227\206" },
-	{ 'a', "\342\226\222" },
-	{ 'f', "\302\260" },
-	{ 'g', "\302\261" },
-	{ 'h', "\342\226\222" },
-	{ 'i', "\342\230\203" },
-	{ 'j', "\342\224\230" },
-	{ 'k', "\342\224\220" },
-	{ 'l', "\342\224\214" },
-	{ 'm', "\342\224\224" },
-	{ 'n', "\342\224\274" },
-	{ 'o', "\342\216\272" },
-	{ 'p', "\342\216\273" },
-	{ 'q', "\342\224\200" },
-	{ 'r', "\342\216\274" },
-	{ 's', "\342\216\275" },
-	{ 't', "\342\224\234" },
-	{ 'u', "\342\224\244" },
-	{ 'v', "\342\224\264" },
-	{ 'w', "\342\224\254" },
-	{ 'x', "\342\224\202" },
-	{ 'y', "\342\211\244" },
-	{ 'z', "\342\211\245" },
-	{ '{', "\317\200" },
-	{ '|', "\342\211\240" },
-	{ '}', "\302\243" },
-	{ '~', "\302\267" }
+	{ '+', "\342\206\222" },	/* arrow pointing right */
+	{ ',', "\342\206\220" },	/* arrow pointing left */
+	{ '-', "\342\206\221" },	/* arrow pointing up */
+	{ '.', "\342\206\223" },	/* arrow pointing down */
+	{ '0', "\342\226\256" },	/* solid square block */
+	{ '`', "\342\227\206" },	/* diamond */
+	{ 'a', "\342\226\222" },	/* checker board (stipple) */
+	{ 'f', "\302\260" },		/* degree symbol */
+	{ 'g', "\302\261" },		/* plus/minus */
+	{ 'h', "\342\226\222" },	/* board of squares */
+	{ 'i', "\342\230\203" },	/* lantern symbol */
+	{ 'j', "\342\224\230" },	/* lower right corner */
+	{ 'k', "\342\224\220" },	/* upper right corner */
+	{ 'l', "\342\224\214" },	/* upper left corner */
+	{ 'm', "\342\224\224" },	/* lower left corner */
+	{ 'n', "\342\224\274" },	/* large plus or crossover */
+	{ 'o', "\342\216\272" },	/* scan line 1 */
+	{ 'p', "\342\216\273" },	/* scan line 3 */
+	{ 'q', "\342\224\200" },	/* horizontal line */
+	{ 'r', "\342\216\274" },	/* scan line 7 */
+	{ 's', "\342\216\275" },	/* scan line 9 */
+	{ 't', "\342\224\234" },	/* tee pointing right */
+	{ 'u', "\342\224\244" },	/* tee pointing left */
+	{ 'v', "\342\224\264" },	/* tee pointing up */
+	{ 'w', "\342\224\254" },	/* tee pointing down */
+	{ 'x', "\342\224\202" },	/* vertical line */
+	{ 'y', "\342\211\244" },	/* less-than-or-equal-to */
+	{ 'z', "\342\211\245" },	/* greater-than-or-equal-to */
+	{ '{', "\317\200" },   		/* greek pi */
+	{ '|', "\342\211\240" },	/* not-equal */
+	{ '}', "\302\243" },		/* UK pound sign */
+	{ '~', "\302\267" }		/* bullet */
 };
 
 int
@@ -81,7 +81,7 @@ tty_acs_get(struct tty *tty, u_char ch)
 	struct tty_acs_entry *entry;
 
 	/* If not a UTF-8 terminal, use the ACS set. */
-	if (!(tty->flags & TTY_UTF8)) {
+	if (tty != NULL && !(tty->flags & TTY_UTF8)) {
 		if (tty->term->acs[ch][0] == '\0')
 			return (NULL);
 		return (&tty->term->acs[ch][0]);
