@@ -291,9 +291,9 @@ utf8_build(void)
 		while (*ptr != NULL) {
 			node = *ptr;
 			if (item->last < node->first)
-				ptr = &(node->left);
+				ptr = &node->left;
 			else if (item->first > node->last)
-				ptr = &(node->right);
+				ptr = &node->right;
 		}
 		*ptr = item;
 	}
